/*********************************************************************
  TinyWanderer_linefollower version 1 

  A line follower program for the TinyWander robot using 2 IR sensors

   Copyright (C) 2011 Doug Paradis

    TinyWanderer Port use:
    ---------------------- 
    PB0 (pin 5): left wheel servo
    PB1 (pin 6): right wheel servo
    PB2/ADC1/INT0 (pin 7): left light sensor
    PB3/ADC3/PCINT3 (pin 2): right light sensor
    PB4 (pin 3): IR LED

  

*********************************************************************/
#define F_CPU 8000000UL  // 8 MHz

#include <util/delay.h> //Includes the delay function
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "tinyWanderer_linefollower.h"



volatile typedef enum {SERVO_INIT, L_SERVO_ON_TIME, L_SERVO_OFF_TIME,
             R_SERVO_ON_TIME, R_SERVO_OFF_TIME, SERVO_PERIOD_REM}
			 servoTimerFlgs; 

volatile typedef enum {
        STATE_INIT,
        STATE_PROC_SENSORS, STATE_GO_FORWARD,
		STATE_GO_LEFT, STATE_GO_RIGHT, STATE_LAST_TURN,
        STATE_FINISHED}
		STATES;


// global variables
volatile int servoTimerFlg;

volatile unsigned char l_servoOnTime;
volatile unsigned char r_servoOnTime;
volatile unsigned char servoOffTime;
volatile unsigned char servoPeriodRemainder;
volatile unsigned int state;

volatile unsigned char ledOffValue;	// phototransistor reading with IR illuminator off
volatile unsigned char ledOnValue;	// phototransistor reading with IR illuminator on

volatile unsigned char ch;	            // ADC channel
volatile unsigned char l_trig_val, r_trig_val;  // servo trigger values

volatile unsigned char lProximityDiff, rProximityDiff;	// difference between LED on & off from proximity phototransistors
volatile unsigned int nextState;
volatile unsigned int endBackupTime;
volatile unsigned char l_cnt = 0, r_cnt = 0;
volatile char last_state_flg = 0;


// function protos
void timerSm(int); 
int state_machine(int);
void state_init();
void state_proc_sensors();
void state_go_forward();
void state_go_left();
void state_go_right();
void timer();
unsigned char measSensor(char, char);
unsigned char readADC(char);


 
/***************************** Main ********************************/
int main(void)
{
   // Port setup
   DDRB = (1<<PB0) | (1<<PB1) | (1<<PB4); // PB0-1,4 as outputs
   PORTB = 0;


   // Timer0
   TCCR0A |= (1<<WGM01);       // Configure Timer0 for CTC
   TIMSK |= (1<<OCIE0A);       // Enable Output Compare Match A interrupt


   

   // Analog Comparator initialization
    ACSR=0x80;      // Analog Comparator: Off

   // ADC setup
   // ADC enable and prescalar = 64 (Clock frequency: 125 kHz)
   ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1);
   // ADC Voltage Reference: Vcc (default) for AREF pin, 0 - VCC
   ADMUX = (1<<ADLAR);  // left justify results 
   DIDR0 = (1<<ADC1D) | (1<<ADC3D);  // disable ADC digital inputs on ch 1 & 3



  // Global interrupts
   sei(); //  Enable global interrupts
   

   // Servos init 
   servoTimerFlg = SERVO_INIT;         // servo_init
   timerSm (servoTimerFlg); 

   l_servoOnTime = LSERVO_STOP;
   r_servoOnTime = RSERVO_STOP;

   state = STATE_INIT;      
   int newState = 0;
 
  /*********** Polling Loop ****************************/

  while( state != STATE_FINISHED )
  {
    switch( state )
    {
      case STATE_INIT:
        state_init();
	  break;

      case STATE_PROC_SENSORS:
        state_proc_sensors();
	  break;

	  case STATE_GO_FORWARD:
        state_go_forward();
	  break;
    
      case STATE_GO_LEFT:
        state_go_left();
	  break;
    
	  case STATE_GO_RIGHT:
        state_go_right();
	  break;
    
	  
    }
 
    newState = state_machine(state);
	state = newState;

  } /******* end of polling loop *******/

} /*************** End of Main ****************/

/******************** subRoutines ***********************************/
/********************************************************************/
int state_machine( int state) 
{
  int newState = 0;

  switch( state )
  {
    case STATE_INIT:
      newState = STATE_PROC_SENSORS;
    break;
 
    case STATE_PROC_SENSORS:
        newState =  nextState;
    break;
  
    case STATE_GO_FORWARD:
          newState = STATE_PROC_SENSORS;
    break;

	case STATE_GO_LEFT:
          newState = STATE_PROC_SENSORS;
    break;

	case STATE_GO_RIGHT:
          newState = STATE_PROC_SENSORS;
    break;

	case STATE_LAST_TURN:
          newState = nextState;
    break;

 
  }

  return(newState);
}
 
/********************************************************************/
void state_init()
{
  // Power on
  char on = 0, off = 1;
  unsigned int l_sum = 0, r_sum = 0;
  unsigned int  l_avg = 0, r_avg = 0;
  char zero_flg = 0;
 
  /**************** Zero Servo Routine ********/
  // To check servo zeros hold robot while not over table and turn on
  // The servos will stay in zero until you turn it back off or put on table
  // I recommend turning off to not confuse the surface calabration

  // trigger values (this could be refined, but works)
  l_trig_val = r_trig_val = 20;  

  while (zero_flg == 0)
  {
     
     // Measure sensors
     ch = 1;  // left sensor ADC ch (PB2-pin 7)
	 ledOffValue = measSensor(ch,off); // Measure Left Sensor with LED off
	 ledOnValue = measSensor(ch,on);  // Measure Left Sensor with LED on
     lProximityDiff = abs(ledOffValue - ledOnValue);

	 ch = 3;   // right sensor ADC ch (PB3-pin 2)
     ledOffValue = measSensor(ch,off); // Measure Left Sensor with LED off
	 ledOnValue = measSensor(ch,on);  // Measure Left Sensor with LED on
     rProximityDiff = abs(ledOffValue - ledOnValue);
  
     // Check to see if sensors are being held off of table  
     if ((lProximityDiff < l_trig_val) && (rProximityDiff < r_trig_val)) // both sensors off table
     {
        l_servoOnTime = LSERVO_STOP;
        r_servoOnTime = RSERVO_STOP;
		zero_flg = 0;
	 }
	 else
	 {
        zero_flg = 1;
	 }
  } 

  // note: turn off robot after zeroing servos and restart 

  /**************** End of Zero Servo Routine ********/

  /**************** Calibrate Surface ********/
  // Calibrate sensor "on table" trigger value
  for (int j = 0; j <=128; j++)  // measure surface 128 times with each sensor
  {
     // Measure sensors
     ch = 1;  // left sensor ADC ch (PB2-pin 7)
	 ledOffValue = measSensor(ch,off); // Measure Left Sensor with LED off
	 ledOnValue = measSensor(ch,on);  // Measure Left Sensor with LED on
     lProximityDiff = abs(ledOffValue - ledOnValue);
	 l_sum += lProximityDiff;

	 ch = 3;   // right sensor ADC ch (PB3-pin 2)
     ledOffValue = measSensor(ch,off); // Measure Left Sensor with LED off
	 ledOnValue = measSensor(ch,on);  // Measure Left Sensor with LED on
     rProximityDiff = abs(ledOffValue - ledOnValue);
	 r_sum += rProximityDiff;
   }
   // calc avg surface read value
   l_avg = (l_sum>>7);          // divide by 128
   r_avg = (r_sum>>7);          // divide by 128

   l_trig_val = (char) (l_avg * 6 / 10);  // trigger value is 40% of avg
   r_trig_val = (char) (r_avg * 6 / 10);   // originally * 4 or 40%
   /**************** End of Calabrate Surface ********/

        
   // Three sec Delay to stablize everything
   _delay_ms(3000); // wait for  3 second  (max is 6.553 sec - resol = 0.1 ms)
   
   // Start Servos
   servoTimerFlg = SERVO_INIT;         // servo_init
   l_servoOnTime = LSERVO_STOP;
   r_servoOnTime = RSERVO_STOP;
   timerSm (servoTimerFlg); 



}
 
/********************************************************************/
void state_proc_sensors()
{
   char off = 1;
   char on = 0;
   char meas_set_flg = 0;

   // wait for SERVO_PERIOD_REM
   while (!(servoTimerFlg == 4));    // wait for SERVO_PERIOD_REM

   // measure sensors only during SERVO_PERIOD_REM
   while ((servoTimerFlg == 4) && (meas_set_flg == 0))
   {

      // Measure sensors
      ch = 1;  // left sensor ADC ch (PB2-pin 7)
      ledOffValue = measSensor(ch,off); // Measure Left Sensor with LED off
	  ledOnValue = measSensor(ch,on);  // Measure Left Sensor with LED on
      lProximityDiff = abs(ledOffValue - ledOnValue);

      ch = 3;   // right sensor ADC ch (PB3-pin 2)
      ledOffValue = measSensor(ch,off); // Measure Right Sensor with LED off
      ledOnValue = measSensor(ch,on);  // Measure Right Sensor with LED on
      rProximityDiff = abs(ledOffValue - ledOnValue);
             
      // analyze sensor results 
      // no cliff - both sensors off line
	  if ((lProximityDiff >= l_trig_val) && (rProximityDiff >= r_trig_val))
      {  
          // no cliff - both servos forward
	      nextState = STATE_GO_FORWARD;
      }
      // left sensor on line
	  else if ((lProximityDiff < l_trig_val) && (rProximityDiff >= r_trig_val))
      {
          // turn to the right (right servo forward fast, left servo forward slow)
	      nextState = STATE_GO_LEFT;
      }
      //  right sensor on line
	  else if ((lProximityDiff >= l_trig_val) && (rProximityDiff < r_trig_val))
      {
          // turn to the left (left servo forward fast, right servo forward slow)
	      nextState = STATE_GO_RIGHT;
      }
	  //  both sensors on line
	  else if ((lProximityDiff >= l_trig_val) && (rProximityDiff >= r_trig_val))
      {
          // keep turning (whatever last turn state)
	      nextState = STATE_LAST_TURN;
      }


	  meas_set_flg = 1;  // set flag after one set of measurements per SERVO_PERIOD_REM
	  while (servoTimerFlg == 4);    // wait for L_SERVO_ON_TIME
      
	  
   }
   

}
 
/********************************************************************/
void state_go_forward()
{
  // no cliff - both servos forward 
  l_servoOnTime = LSERVO_FWD;
  r_servoOnTime = RSERVO_FWD;

  last_state_flg = CLEAR;

  l_cnt = 0;
  r_cnt = 0;

  
}

/********************************************************************/
void state_go_left()  // left sensor on line
{
  // turn to the left (right servo forward fast, left servo stop)
  l_servoOnTime = LSERVO_FWD_SLO;
  r_servoOnTime = RSERVO_FWD;

  last_state_flg = LEFT;
  l_cnt++;
  r_cnt = 0;

  if (l_cnt > 4)
  {
     l_servoOnTime = LSERVO_TURN_SLO;
  }
}

/********************************************************************/
void state_go_right()     // left sensor on line
{
  // turn to the right (left servo forward fast, right servo stop) 
  l_servoOnTime = LSERVO_FWD;
  r_servoOnTime = RSERVO_FWD_SLO;

  last_state_flg = RIGHT;
  r_cnt++;
  l_cnt = 0;

  if (r_cnt > 4)
  {
     r_servoOnTime = RSERVO_TURN_SLO;
  }

  
}

 
/********************************************************************/
void state_last_turn()     // both sensors on line
// keep going into last turn

{
 // determine last turn state
 switch( last_state_flg)
  {
    case RIGHT:
      nextState = STATE_GO_RIGHT;
    break;
 
    case LEFT:
        nextState = STATE_GO_LEFT;
    break;

  
  }
  
  
}
/********************************************************************/
unsigned char measSensor(char ch, char led_pwr)
{
     char adc_result;
   
     if (led_pwr == 1)
	 {
	    PORTB &= ~(1<<PB4);			// turn IR LEDs off (low light = high output)
		//  add delay -- photo xtr is slow, delay allows max signal value
        _delay_ms(1.5); // wait for 1.5 milliseconds
     }
	 else 
	 {
	    PORTB |= (1<<PB4);			// turn IR LEDs on  (light makes output low)
		//  add delay -- photo xtr is slow, delay allows max signal value
        _delay_ms(1.5); // wait for 1.5 milliseconds 
	   
     } 

     adc_result = readADC(ch); 
 
	
	 

	 return(adc_result);
}


/********************************************************************/
unsigned char readADC(char ch)
{
    if (ch == 1)
	{
	   ADMUX &= ~(0x13);
	   ADMUX |= 0x11;    // Left Sensor
    }
	if (ch == 3)
	{
	   ADMUX &= ~(0x11);
	   ADMUX |= 0x13;    // Right Sensor
     }

	

	// Start conversion
	ADCSRA |= (1<<ADSC);

	// Wait for conversion to complete
	while (!(ADCSRA & (1<<ADIF)));

	// Clear ADIF by writing one to it
	// Per datasheet writing one clears bit in register
	ADCSRA |= (1<<ADIF);

    return(ADCH);   // 8 bit resolution
}

/********************************************************************/
void timerSm(int servoTimerFlg)
{
   switch(servoTimerFlg)
   {
      
      case SERVO_INIT:              // servo_init
	    TCCR0B &= ~(1<<CS01);             // prescalar set to divide by 1024 (255 ~= 32ms)
        TCCR0B = (1<<CS00) | (1<<CS02);   // prescalar set to divide by 1024 (255 ~= 32ms)
		OCR0A = 160;                      // wait one servo cycle before starting 160 ~= 20mS 
      break;                             
	    
	  
      case L_SERVO_ON_TIME:            //l_servo_on_time: 
	     PORTB |= (1<<PB0);     //left
		 TCCR0B &= ~(1<<CS02);              // reset prescalar
         TCCR0B = (1<<CS00) | (1<<CS01);   // prescaler set to divide by 64 (255 ~= 2ms)
		 OCR0A = l_servoOnTime;          // Set Output Compare Register A compare value (191 ~1.5ms) 
       break;                            
      
	   
	   case L_SERVO_OFF_TIME:             //l_servo_off_time:
	     PORTB &= ~(1<<PB0);    //left
		 servoOffTime = 64 + (255 - l_servoOnTime);  // rest of servo 2.5ms period
		 OCR0A = servoOffTime;         // Set Output Compare Register A compare value
       break;

	   
       case R_SERVO_ON_TIME:            //r_servo_on_time:
         PORTB |= (1<<PB1);     //right
         OCR0A = r_servoOnTime;          // Set Output Compare Register A compare value (191 ~1.5ms)
       break;                           
      
	   
	   case R_SERVO_OFF_TIME:            //r_servo_off_time:
         PORTB &= ~(1<<PB1);     //right
		 servoOffTime = 64 + (255 - r_servoOnTime);  // rest of servo 2.5ms period
		 OCR0A = servoOffTime;         // Set Output Compare Register A compare value
       break;


       case SERVO_PERIOD_REM:            //servo_period_rem:
		 GTCCR = (1<<PSR0) | (1<<TSM);
		 TCCR0B &= ~(1<<CS01);               // prescaler set to divide by 1024 (255 ~= 32ms)
	     TCCR0B = (1<<CS00) | (1<<CS02);   // prescaler set to divide by 1024 (255 ~= 32ms)
		 GTCCR &= ~(1<<TSM);
		 servoPeriodRemainder = 120;       // 120 ~= 15mS, 84 ~= 10.5 mS - forcing short servo period
         OCR0A = servoPeriodRemainder;     // Set Output Compare Register A compare value
       break;
    }

} 



/******************** ISR Routine ***********************************/
/********************************************************************/
ISR(TIM0_COMPA_vect)
// ISR(TIM0_OVF_vect)
{
  timerSm(servoTimerFlg);
  if (servoTimerFlg <= 4) 
  {
    servoTimerFlg++;
  }
  else
  {
     servoTimerFlg = 1;

  }

} 

/********************************************************************/
